
/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 5
 */

#include "squidwall.h"
#include "parse_url.h"
#include "read_acls.h"

/* from main.c */
extern stralloc request;
extern pthread_mutex_t request_mutex;

void handle_request(request_t *r)
{
	int flags, todo, state, j;
	unsigned int i;
	stralloc sa, out;
	
	stralloc_init(&sa);
	stralloc_init(&out);
	if (!stralloc_copy(&sa, &request)) die_nomem();

	debug(1,5, "UNlooking the request variable ...");
	pthread_mutex_unlock(&request_mutex);

	read_acls(r);

	stralloc_init(&r->ip);
	stralloc_init(&r->url);
	stralloc_init(&r->host);
	stralloc_init(&r->user);
	stralloc_init(&r->method);

	stralloc_init(&r->redirsite);
	stralloc_init(&r->clamav);
	stralloc_init(&r->filter);
	stralloc_init(&r->virus);

	stralloc_init(&r->url_host);
	stralloc_init(&r->url_path);

	stralloc_init(&r->pfilter_matches);

	flags=ACL_NONE; todo=TODO_NONE; state=0;

	/**
	 * format_v1: "url ip/host user method"
	 * format_v2: "nr url ip/host user method urlgroup"
	 *
	 * sample1: http://www.mcmilk.de/ 192.168.100.10/aspire mcmilk GET
	 * sample2: www.blabla.de:443 192.168.100.6/p28tr - CONNECT
	 *
	 * state=0 url=http://www.mcmilk.de/
	 * state=1 ip=192.168.100.10
	 * state=2 host=aspire
	 * state=3 user=root
	 * state=4 method=GET
	 */

	r->squid_url=sa.s;

	/**
	 * look for url_rewrite_concurrency
	 */
	j = scan_uint(sa.s, &r->tid);
	if (j) {
		j+=1;
		set(flags, ACL_MULTI);
		debug(5,3, "squid request thread: %d", r->tid);
		r->squid_url=sa.s+j;
	}

	for (i=j; i<=sa.len; i++) {
		switch (state) {
		case 0:
		if (sa.s[i] == ' ') {
			sa.s[i]=0; state=1; r->squid_ip=sa.s+i+1;
		}
		break;

		case 1:
		if (sa.s[i] == '/') {
			sa.s[i]=0; state=2; r->squid_host=sa.s+i+1;
		}
		break;

		case 2:
		if (sa.s[i] == ' ') {
			sa.s[i]=0; state=3; r->squid_user=sa.s+i+1;
		}
		break;

		case 3:
		if (sa.s[i] == ' ') {
			sa.s[i]=0; state=4; r->squid_method=sa.s+i+1;
		}
		break;

		case 4:
		if (sa.s[i] == ' ') {
			sa.s[i]=0; state=5;
		}
		break;

		} /* switch() */
	}


	if (state != 4 && state != 5) {
		debug(5,1, "error: incorrect input from squid!");
		goto out;
	}

	debug(5,1, "INPUT: ip(%s) url(%s) host(%s) user(%s) method(%s)!",
		r->squid_ip, r->squid_url, r->squid_host, r->squid_user, r->squid_method);

	/* seperate host/path from the url */
	//parse_url(r);
	//debug(5,3, "parse_url(): host(%s)", r->url_host.s);
	//debug(5,3, "parse_url(): path(%s)", r->url_path.s);

	state=0;
	if (str_start(r->squid_method, "HEAD")) state=1;
	if (str_start(r->squid_method, "POST")) state=1;
	if (str_start(r->squid_method, "GET")) state=1;

	if (state != 1) {
		/* direkt redirecting CONNECT will cause failures
		 * -> we add a "302:" prefix later
		 */
		set(flags, ACL_REDIR);
	}

	/* squid gives us urlencoded things - we decode them first! */
	if (!scan_to_sa(scan_urlencoded,  r->squid_ip, &r->ip)) die_nomem();
	if (!scan_to_sa(scan_urlencoded2, r->squid_url, &r->url)) die_nomem();
	if (!scan_to_sa(scan_urlencoded,  r->squid_host, &r->host)) die_nomem();
	if (!scan_to_sa(scan_urlencoded2, r->squid_user, &r->user)) die_nomem();
	if (!scan_to_sa(scan_urlencoded,  r->squid_method, &r->method)) die_nomem();

	r->ip.s[r->ip.len]=0;
	r->url.s[r->url.len]=0;
	r->host.s[r->host.len]=0;
	r->user.s[r->user.len]=0;
	r->method.s[r->method.len]=0;

	debug(5,1, "INPUT: ip(%s) url(%s) host(%s) user(%s) method(%s)!",
		r->squid_ip, r->squid_url, r->squid_host, r->squid_user, r->squid_method);

	/* we check in this order:
	 *
	 *
	 * 1  is the url http://squidwall -> redirect to webinterface
	 * 2  check for redirect sites -> redirect if needed
	 * 3. check the acl's (global, user, host, ip)
	 * 4. look for ad's / banners
	 * 5. look for whitelisted url's
	 * 6. look for blacklisted url's
	 * 7. check for bad content
	 *    - pfilter, cfilter, crfilter
	 *    - antivirus
	 *    - bcfilter (banner via content regex filter)
	 *    - learning / classifying
	 */

	/* the internet explorer will not redirect per default
	 * - it will just try to resolve the host 'squidwall' and then
	 *   redirect to your favorite search engine ...
	 * - just use http://squidwall.lala.de/
	 */
	if (str_start(r->squid_url, "http://squidwall")) {
		redir_site(r, REDIR_TYPE_CONFIG);
		set(flags, ACL_REDIR);
		goto out;
	}
	goto out;

	/* just pass through */
	//if (str_start(r->squid_url, redir_safesite)) {
	//	goto out;
	//}

	/* check, if the url matches some user defined redirect site */
	//if (check_redirect(r)) {
	//	/* stralloc redirsite is set in the function */
	//	goto out;
	//}

	/* recheck the filter definitions and acl's */
	//read_acls(r);

	/* check, if it was cached */
	//if (memcache_lookup(r) == MEMCACHE_HIT) {
	//	set(flags, ACL_MEMHIT);
	//	goto out;
	//}

	/* get the acl's now correctly set */
	flags |= setup_acls(r, check_user(r, &todo), check_host(r, &todo),
		check_globals(r, &todo));
	//debug_flags("nach setup_acls():", flags);
	//debug_todo("nach setup_acls():", todo);

	/* check for banner */
	if (isset(flags, ACL_FILTER)) {
		flags |= check_banner(r);
	}
	if (isset(flags, ACL_DENY)) {
		redir_site(r, REDIR_TYPE_BANNER);
		goto out;
	}
	//debug_flags("nach check_banner():", flags);
	//debug_todo("nach check_banner():", todo);
	
	/* check for whitelisted url's */
	if (isset(flags, ACL_FILTER))
	if (check_wlist(r) == ACL_ALLOW) {
		goto out;
	}
	//debug_flags("nach check_wlist():", flags);
	//debug_todo("nach check_wlist():", todo);

	/* check for blacklisted url's */
	if (isset(flags, ACL_FILTER)) {
		flags |= check_blist(r);
	}
	if (isset(flags, ACL_DENY)) {
		redir_site(r, REDIR_TYPE_URL);
		goto out;
	}
	//debug_flags("nach check_blist():", flags);
	//debug_todo("nach check_blist():", todo);

	/* last step: scan / learn / classify the content */
	if (isset(flags, ACL_FILTER)) {
		check_cont(r, todo);
	}
	//debug_flags("nach check_cont():", flags);
	//debug_todo("nach check_cont():", todo);

	out:
	//log_flush();

	/* this is currently no real performance improvement, we
	 * need threading for that ... TODO
	 */
	if (isset(flags, ACL_MULTI)) {
		if (!stralloc_cat_uint(&out, r->tid)) die_nomem();
		if (!stralloc_cats(&out, " ")) die_nomem();
	}

	/* '302:' prefix for methods, other than: HEAD, GET, POST */
	if (isset(flags, ACL_REDIR)) {
		if (r->redirsite.len)
			if (!stralloc_cats(&out, "302:")) die_nomem();
	}

	/* cache this redirsite */
	#if 0
	if (!isset(flags, ACL_MEMHIT)) {
		memcache_insert(r);
	} else {
		debug(5,3, "squidwall: memcache hit on (%s)", r->squid_url);
	}
	#endif

	if (!stralloc_cat(&out, &r->redirsite)) die_nomem();
	if (!stralloc_cats(&out, "\n")) die_nomem();
	out.s[out.len]=0;

	/* give squid an empty newline '\n' or the whole redirsite */
	io_wantwrite(1); io_wait(); io_dontwantwrite(1);
	j = write(1, out.s, out.len); /* no perfect errorchecking needed! */
	if (j == -1 && errno == EPIPE) {
		panic("squid has closed the connection!");
		exit(111);
	}

	debug(1,1,"write(1,...)");
#if LOGTODO
	if (isset(logtodo, LOG_SQUID_REQUEST)) {
		log_headline("request");
		if (isset(flags, ACL_MULTI)) {
			log_ulong(r->squid_id); log_s(" ");
		}
		log_s(r->squid_url); log_s(" ");
		log_s(r->squid_ip); log_s("/");
		log_s(r->squid_host); log_s(" ");
		log_s(r->squid_user); log_s(" ");
		log_s(r->squid_method); log_s("\n");
	}

	if (isset(logtodo, LOG_SQUIDWALL_RESPONSE)) {
		out.s[out.len-1]=0; /* remove \n */
		log_headline("answer");
		log_s(out.s); log_s("\n");
	}
#endif

#if 0
	{
	struct timespec t;

		t.tv_sec=0; t.tv_nsec=350;
		nanosleep(&t,&t); /* let the thread start */
	}
#endif

	stralloc_free(&sa);
	stralloc_free(&out);

	stralloc_free(&r->ip);
	stralloc_free(&r->url);
	stralloc_free(&r->host);
	stralloc_free(&r->user);
	stralloc_free(&r->method);
	stralloc_free(&r->redirsite);

	stralloc_free(&r->url_host);
	stralloc_free(&r->url_path);

	return;
}
